@@ -1,6 +1,6 @@
 /*
- * Copyright (C) 2003 Nikos Mavroyanopoulos
  * Copyright (C) 2004 Free Software Foundation
+ * Copyright (C) 2003 Nikos Mavroyanopoulos
  *
  * This file is part of GNUTLS.
  *
@@ -246,12 +246,16 @@ gnutls_x509_crt generate_certificate( gnutls_x509_privkey *ret_key,
 		fprintf(stderr, "Please enter the details of the certificate's distinguished name. "
 		"Just press enter to ignore a field.\n");
 
+		/* set the DN.
+		 */
 		get_country_crt_set( crt);
 		get_organization_crt_set(crt);
 		get_unit_crt_set( crt);
 		get_locality_crt_set( crt);
 		get_state_crt_set( crt);
 		get_cn_crt_set( crt);
+		get_uid_crt_set( crt);
+		get_oid_crt_set( crt);
 	
 		if (!batch) fprintf(stderr, "This field should not be used in new certificates.\n");
 
@@ -319,15 +323,23 @@ gnutls_x509_crt generate_certificate( gnutls_x509_privkey *ret_key,
 
 	server = get_tls_server_status();
 	if (server != 0) {
+		result = 0;
+
 		str = get_dns_name();
 		if (str != NULL) {
 			result = gnutls_x509_crt_set_subject_alternative_name( crt, GNUTLS_SAN_DNSNAME, str);
-			if (result < 0) {
-				fprintf(stderr, "subject_alt_name: %s\n", gnutls_strerror(result));
-				exit(1);
+		} else {
+			str = get_ip_addr();
+			if (str != NULL) {
+				result = gnutls_x509_crt_set_subject_alternative_name( crt, GNUTLS_SAN_IPADDRESS, str);
 			}
 		}
 
+		if (result < 0) {
+			fprintf(stderr, "subject_alt_name: %s\n", gnutls_strerror(result));
+			exit(1);
+		}
+
 		result = gnutls_x509_crt_set_key_purpose_oid( crt, GNUTLS_KP_TLS_WWW_SERVER, 0);
 		if (result < 0) {
 			fprintf(stderr, "key_kp: %s\n", gnutls_strerror(result));
@@ -855,7 +867,7 @@ static void print_certificate_info( gnutls_x509_crt crt, FILE* out, unsigned int
 {
 	int ret;
 	unsigned int i, indx, j, version;
-	unsigned int critical, key_usage;
+	unsigned int critical, key_usage, bits;
 	time_t tim;
 	char serial[40];
 	size_t serial_size = sizeof(serial), dn_size, size;
@@ -926,13 +938,14 @@ static void print_certificate_info( gnutls_x509_crt crt, FILE* out, unsigned int
 	/* Public key algorithm
 	 */
 	fprintf(out, "Subject Public Key Info:\n");
-	ret = gnutls_x509_crt_get_pk_algorithm(crt, NULL);
+	ret = gnutls_x509_crt_get_pk_algorithm(crt, &bits);
 	fprintf(out, "\tPublic Key Algorithm: ");
 
 	cprint = gnutls_pk_algorithm_get_name( ret);
 	if (cprint == NULL) cprint = UNKNOWN;
-	fprintf(out,  "%s\n", cprint);
-
+	fprintf(out,  "%s", cprint);
+	if (bits) fprintf(out,  " (%u bits)", bits);
+	fprintf(out, "\n");
 
 	if (version >= 3)
 		fprintf(out, "\nX.509 Extensions:\n");
@@ -1535,7 +1548,7 @@ int size;
 
 	crt = load_cert_list( mand, &size);
 
-	return crt[0];
+	return crt ? crt[0] : NULL;
 }
 
 #define MAX_CERTS 256
@@ -1631,12 +1644,16 @@ void generate_request(void)
 	 */
 	key = generate_private_key_int();
 
+	/* Set the DN.
+	 */
 	get_country_crq_set( crq);
 	get_organization_crq_set(crq);
 	get_unit_crq_set( crq);
 	get_locality_crq_set( crq);
 	get_state_crq_set( crq);
 	get_cn_crq_set( crq);
+	get_uid_crq_set( crq);
+	get_oid_crq_set( crq);
 
 	ret = gnutls_x509_crq_set_version( crq, 1);
 	if (ret < 0) {
@@ -2220,9 +2237,9 @@ void print_bag_data(gnutls_pkcs12_bag bag)
 {
 int result;
 int count, i, type;
-gnutls_const_datum cdata;
+gnutls_datum cdata;
 const char* str;
-gnutls_datum out, data;
+gnutls_datum out;
 
 	count = gnutls_pkcs12_bag_get_count( bag);
 	if (count < 0) {
@@ -2266,14 +2283,8 @@ gnutls_datum out, data;
 			str = NULL;
 		}		
 	
-		/* we have to cast gnutls_const_datum to a
-		 * plain datum.
-		 */
-		data.data = (unsigned char*)cdata.data;
-		data.size = cdata.size;
-
 		if (str != NULL) {
-			gnutls_pem_base64_encode_alloc( str, &data, &out);
+			gnutls_pem_base64_encode_alloc( str, &cdata, &out);
 			fprintf( outfile, "%s\n", out.data);
 		
 			gnutls_free(out.data);
